@@ -13,6 +13,12 @@ import UIKit
 import AVFoundation
 import MediaPlayer
 
+enum PlayerSeekState {
+    case `default`
+    case forward
+    case backward
+}
+
 enum PlayerPanType {
     case none
     case brightness
@@ -67,6 +73,8 @@ class PlayerViewController: UIViewController {
     var alertController: UIAlertController?
     var seekForwardBy: Int = 0
     var seekBackwardBy: Int = 0
+    var numberOfTapSeek: Int = 0
+    var previousSeekState: PlayerSeekState = .default
 
     lazy var statusLabel: VLCStatusLabel = {
         var statusLabel = VLCStatusLabel()
@@ -243,6 +251,8 @@ class PlayerViewController: UIViewController {
 
     private var fov: CGFloat = 0
 
+    private var viewTranslation: CGPoint = CGPoint(x: 0, y: 0)
+
     private lazy var deviceMotion: DeviceMotion = {
         let deviceMotion = DeviceMotion()
         deviceMotion.delegate = self
@@ -253,7 +263,7 @@ class PlayerViewController: UIViewController {
 
     private let screenPixelSize = CGSize(width: UIScreen.main.bounds.width,
                                          height: UIScreen.main.bounds.height)
-
+    
     // MARK: - Gestures
 
     lazy var panRecognizer: UIPanGestureRecognizer = {
@@ -290,6 +300,19 @@ class PlayerViewController: UIViewController {
         return rightSwipeRecognizer
     }()
 
+    lazy var minimizeGestureRecognizer: UIPanGestureRecognizer = {
+        let minimizeGestureRecognizer = UIPanGestureRecognizer(target: self,
+                                                               action: #selector(handleMinimizeGesture(_:)))
+        return minimizeGestureRecognizer
+    }()
+
+    lazy var doubleTapGestureRecognizer: UITapGestureRecognizer = {
+        let doubleTapGestureRecognizer = UITapGestureRecognizer(target: self,
+                                                                action: #selector(handleDoubleTapGesture(_:)))
+        doubleTapGestureRecognizer.numberOfTapsRequired = 2
+        return doubleTapGestureRecognizer
+    }()
+
     // MARK: - Init
 
     @objc init(mediaLibraryService: MediaLibraryService, rendererDiscovererManager: VLCRendererDiscovererManager, playerController: PlayerController) {
@@ -298,11 +321,17 @@ class PlayerViewController: UIViewController {
         self.playerController = playerController
         super.init(nibName: nil, bundle: nil)
         mediaNavigationBar.chromeCastButton = rendererButton
+        mediaNavigationBar.addGestureRecognizer(minimizeGestureRecognizer)
     }
 
     required init?(coder: NSCoder) {
         fatalError("init(coder:) has not been implemented")
     }
+    
+    override func viewDidLoad() {
+        setupObservers()
+        hideSystemVolumeInfo()
+    }
 
     override func viewWillAppear(_ animated: Bool) {
         setupRendererDiscovererManager()
@@ -316,6 +345,13 @@ class PlayerViewController: UIViewController {
         }
 
         mediaNavigationBar.updateDeviceButton(with: image, color: color)
+
+        if playerController.isRememberStateEnabled {
+            playbackService.isShuffleMode = playerController.isShuffleEnabled
+            playbackService.repeatMode = playerController.isRepeatEnabled
+        }
+
+        view.transform = .identity
     }
 
     // MARK: - Public methods
@@ -349,12 +385,56 @@ class PlayerViewController: UIViewController {
         pinchRecognizer.isEnabled = !disable
         leftSwipeRecognizer.isEnabled = !disable
         rightSwipeRecognizer.isEnabled = !disable
+        doubleTapGestureRecognizer.isEnabled = !disable
+    }
+    
+    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey: Any]?, context: UnsafeMutableRawPointer?) {
+        // We're observing outputVolume to handle volume changes from physical volume buttons
+        // To processd properly we have to check we're not interacting with UI controls or gesture
+        if  keyPath == "outputVolume" &&
+            !volumeControlView.isBeingTouched && // Check we're not interacting with volume slider
+            !isGestureActive && // Check if the slider did not just changed value
+            currentPanType != .volume { // Check we're not doing pan gestures for volume
+            let appearAnimations = { [volumeControlView] in
+                volumeControlView.alpha = 1
+            }
+            let disappearAnimations = { [volumeControlView] in
+                volumeControlView.alpha = 0
+            }
+            UIView.animate(withDuration:0.2, delay: 0,
+                           options: .beginFromCurrentState,
+                           animations:appearAnimations, completion:nil)
+            UIView.animate(withDuration: 0.2, delay: 1,
+                           options: [],
+                           animations:disappearAnimations, completion:nil)
+            self.volumeControlView.updateIcon(level: AVAudioSession.sharedInstance().outputVolume)
+        }
     }
 
     func changeOutputView(to output: UIView?) {
         // Change the content output view if necessary according to the type of the player.
     }
 
+    func minimizePlayer() {
+        // Minimize the player
+    }
+
+    func updateShuffleState() {
+        playbackService.isShuffleMode = !playbackService.isShuffleMode
+
+        if playerController.isRememberStateEnabled {
+            UserDefaults.standard.setValue(playbackService.isShuffleMode, forKey: kVLCPlayerIsShuffleEnabled)
+        }
+    }
+
+    func updateRepeatMode() {
+        playbackService.toggleRepeatMode()
+
+        if playerController.isRememberStateEnabled {
+            UserDefaults.standard.setValue(playbackService.repeatMode.rawValue, forKey: kVLCPlayerIsRepeatEnabled)
+        }
+    }
+
     // MARK: - Private methods
 
     private func jumpBackwards(_ interval: Int = 10) {
@@ -365,6 +445,27 @@ class PlayerViewController: UIViewController {
         playbackService.jumpForward(Int32(interval))
     }
 
+    private func executeSeekFromTap() {
+        var hudString: String = ""
+        let currentSeek = numberOfTapSeek > 0 ? seekForwardBy : seekBackwardBy
+        let seekDuration: Int = numberOfTapSeek * currentSeek
+
+        if seekDuration > 0 {
+            hudString = "⇒ "
+            jumpForwards(currentSeek)
+            previousSeekState = .forward
+        } else {
+            hudString = "⇐ "
+            jumpBackwards(currentSeek)
+            previousSeekState = .backward
+        }
+
+        // Convert the time in seconds into milliseconds in order to the get the right VLCTime value.
+        let duration: VLCTime = VLCTime(number: NSNumber(value: abs(seekDuration) * 1000))
+        hudString.append(duration.stringValue)
+        statusLabel.showStatusMessage(hudString)
+    }
+
     private func showIcon(button: UIButton) {
         UIView.animate(withDuration: 0.5, animations: {
             button.isHidden = false
@@ -470,6 +571,16 @@ class PlayerViewController: UIViewController {
         rendererDiscovererManager.presentingViewController = self
         rendererDiscovererManager.delegate = self
     }
+    
+    private func hideSystemVolumeInfo() {
+        volumeView.alpha = 0.00001
+        view.addSubview(volumeView)
+    }
+    
+    private func setupObservers() {
+        try? AVAudioSession.sharedInstance().setActive(true)
+        AVAudioSession.sharedInstance().addObserver(self, forKeyPath: "outputVolume", options: NSKeyValueObservingOptions.new, context: nil)
+    }
 
     // MARK: - Gesture handlers
 
@@ -656,6 +767,51 @@ class PlayerViewController: UIViewController {
             statusLabel.showStatusMessage(hudString)
         }
     }
+
+    @objc private func handleMinimizeGesture(_ sender: UIPanGestureRecognizer) {
+        switch sender.state {
+        case .changed:
+            viewTranslation = sender.translation(in: view)
+            if viewTranslation.y < 0 {
+                return
+            }
+
+            UIView.animate(withDuration: 0.5,
+                           delay: 0,
+                           usingSpringWithDamping: 0.7,
+                           initialSpringVelocity: 1,
+                           options: .curveEaseOut,
+                           animations: {
+                self.view.transform = CGAffineTransform(translationX: 0, y: self.viewTranslation.y)
+            })
+
+            break
+        case .ended:
+            let halfHeight = view.frame.height / 2
+            if viewTranslation.y < halfHeight {
+                UIView.animate(withDuration: 0.5,
+                               delay: 0,
+                               usingSpringWithDamping: 0.7,
+                               initialSpringVelocity: 1,
+                               options: .curveEaseOut,
+                               animations: {
+                    self.view.transform = .identity
+                })
+            } else {
+                minimizePlayer()
+            }
+
+            break
+        default:
+            break
+        }
+    }
+
+    @objc func handleDoubleTapGesture(_ sender: UITapGestureRecognizer) {
+        // CHECK THE TAP LOCATION
+
+        executeSeekFromTap()
+    }
 }
 
 // MARK: - VLCPlaybackServiceDelegate
@@ -839,17 +995,14 @@ extension PlayerViewController: MediaMoreOptionsActionSheetDelegate {
     }
 
     func mediaMoreOptionsActionSheetDidToggleShuffle(_ mediaMoreOptionsActionSheet: MediaMoreOptionsActionSheet) {
-        playbackService.isShuffleMode = !playbackService.isShuffleMode
-
-        if playerController.isRememberStateEnabled {
-            UserDefaults.standard.setValue(playbackService.isShuffleMode, forKey: kVLCPlayerIsShuffleEnabled)
-        }
+        updateShuffleState()
 
         mediaMoreOptionsActionSheet.collectionView.reloadData()
     }
 
     func mediaMoreOptionsActionSheetDidTapRepeat(_ mediaMoreOptionsActionSheet: MediaMoreOptionsActionSheet) {
-        playbackService.toggleRepeatMode()
+        updateRepeatMode()
+
         mediaMoreOptionsActionSheet.collectionView.reloadData()
     }
 }
